@@ -1,15 +1,17 @@
+#
+# This file is part of pyspex
+#
+# https://github.com/rmvanhees/pyspex.git
+#
+# Copyright (c) 2019-2022 SRON - Netherlands Institute for Space Research
+#    All Rights Reserved
+#
+# License:  BSD-3-Clause
 """
-This file is part of pyspex
-
-https://github.com/rmvanhees/pyspex.git
-
-Class to read SPEXone ICU packages (version 2)
-
-Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
-   All Rights Reserved
-
-License:  BSD-3-Clause
+Contains the class `CCSDSio` to read SPEXone telemetry packets.
 """
+__all__ = ['CCSDSio']
+
 from pathlib import Path
 
 import numpy as np
@@ -36,7 +38,7 @@ from pyspex.lib.tmtc_def import tmtc_dtype
 #  - Packet length  (16 bits): size of packet data in bytes (always odd)
 #                              (secondary header + User data) - 1
 #  - Packet timestamp:         Secondary header
-#                   (32 bits): seconds (TAI)
+#                   (32 bits): seconds (1-1-1970 UTC)
 #                   (16 bits): sub-seconds (1/2 ** 16)
 
 HDR_DTYPE = np.dtype([
@@ -54,53 +56,22 @@ TIME_DTYPE = np.dtype([
 
 SCIHK_DTYPE = tmtc_dtype(0x350)
 
-# - local functions --------------------------------
+# - Error messages ------------------------
+MSG_SKIP_FRAME = "[WARNING]: rejected a frame because it's incomplete"
+MSG_INVALID_APID = \
+    '[WARNING]: found one or more telemetry packages with an invalid APID'
+MSG_CORRUPT_APID = 'corrupted segements - detected APID 1 after <> 2'
+MSG_CORRUPT_FRAME = 'corrupted segements - previous frame not closed'
 
 
 # - class CCSDSio -------------------------
 class CCSDSio:
-    """
-    Read SPEXone telemetry packets.
+    """Read SPEXone telemetry packets.
 
-    Attributes
+    Parameters
     ----------
-    found_invalid_apid :  bool
-    file_list :  iter
-    fp :  file pointer
-
-    Methods
-    -------
-    version_no
-       Returns CCSDS version number.
-    type_indicator
-       Returns type of telemetry packet.
-    secnd_hdr_flag
-       Returns flag indicating presence of a secondary header.
-    ap_id
-       Returns SPEXone ApID.
-    grouping_flag
-       Returns grouping flag.
-    sequence_count
-       Returns sequence counter, rollover to zero at 0x3FFF.
-    packet_length
-       Returns size of packet data in bytes.
-
-    close()
-       Close resources.
-    open_next_file()
-       Open next file from file_list.
-    fix_dem_hk24(sci_hk)
-       Correct 32-bit integers in the DemHK which originate from
-       24-bit integers in the detector register values.
-    fix_sci_hk24(sci_hk)
-       Correct 32-bit integers in the Science_HK which originate from
-       24-bit integers in the detector register values.
-    read_packet()
-       Read next telemetry packet.
-    select_tm(packets, ap_id)
-       Select NomHK telemetry packages
-    science_tm(packets_in)
-       Combine segmented Science telemetry packages.
+    file_list: list of strings
+        list of file-names, where each file contains parts of a measurement
 
     Notes
     -----
@@ -116,8 +87,13 @@ class CCSDSio:
 
     Doc: TMTC handbook (SPX1-TN-005), issue 12, 15-May-2020
 
+    The files with science and telemetry data needs to be in chronological
+    order. However, you may mix science and housekeeping data as long as
+    science data are chronological and housekeeping data are chronological.
+
     Examples
     --------
+
     >>> packets = ()
     >>> with CCSDSio(file_list) as ccsds:
     >>>     while True:
@@ -132,35 +108,24 @@ class CCSDSio:
     >>>     # combine segmented Science packages
     >>>     science_tm = ccsds.group_tm(packets)
     >>>     # now you may want to collect the engineering packages
+
     """
-    def __init__(self, file_list: str) -> None:
-        """
-        Initialize access to a SPEXone Level-0 product (CCSDS format)
-
-        Parameters
-        ----------
-        file_list: list of strings
-           list of file-names, where each file contains parts of a measurement
-
-        Notes
-        -----
-        The files with science and telemetry data needs to be in chronological
-        order. However, you may mix science and housekeeping data as long as
-        science data are chronological and housekeeping data are chronological.
+    def __init__(self, file_list: list) -> None:
+        """Initialize access to a SPEXone Level-0 product (CCSDS format).
         """
         # initialize class attributes
-        self.__hdr = None
-        self.found_invalid_apid = False
         self.file_list = iter(file_list)
+        self.found_invalid_apid = False
+        self.__hdr = None
         self.fp = None
 
-        self.open_next_file()
+        if file_list:
+            self.open_next_file()
 
     def __repr__(self) -> str:
-        return (f'{self.version_no:03d} {self.type_indicator}'
-                f' {self.secnd_hdr_flag} 0x{self.ap_id:x} {self.grouping_flag}'
-                f' {self.sequence_count:5d} {self.packet_length:5d}'
-                f' {self.fp.tell():9d}')
+        return (f'{self.version_no:03d} 0x{self.ap_id:x}'
+                f' {self.grouping_flag} {self.sequence_count:5d}'
+                f' {self.packet_length:5d} {self.fp.tell():9d}')
 
     def __iter__(self):
         for attr in sorted(self.__dict__):
@@ -168,34 +133,29 @@ class CCSDSio:
                 yield attr
 
     def __enter__(self):
-        """
-        method called to initiate the context manager
+        """Method called to initiate the context manager.
         """
         return self
 
     def __exit__(self, exc_type, exc_value, traceback) -> bool:
-        """
-        method called when exiting the context manager
+        """Method called when exiting the context manager.
         """
         self.close()
         return False  # any exception is raised by the with statement.
 
     def close(self) -> None:
-        """
-        Close resources
+        """Close resources.
         """
         if self.fp is not None:
             if self.found_invalid_apid:
-                print('[WARNING]: found one or more telemetry packages'
-                      ' with an invalid APID')
+                print(MSG_INVALID_APID)
             self.found_invalid_apid = False
             self.fp.close()
 
     # ---------- define some class properties ----------
     @property
     def version_no(self) -> int:
-        """
-        Returns CCSDS version number
+        """Returns CCSDS version number.
         """
         if self.__hdr is None:
             return None
@@ -204,8 +164,7 @@ class CCSDSio:
 
     @property
     def type_indicator(self) -> int:
-        """
-        Returns type of telemetry packet
+        """Returns type of telemetry packet.
         """
         if self.__hdr is None:
             return None
@@ -214,8 +173,7 @@ class CCSDSio:
 
     @property
     def secnd_hdr_flag(self) -> bool:
-        """
-        Returns flag indicating presence of a secondary header
+        """Returns flag indicating presence of a secondary header.
         """
         if self.__hdr is None:
             return None
@@ -224,8 +182,7 @@ class CCSDSio:
 
     @property
     def ap_id(self) -> int:
-        """
-        Returns SPEXone ApID
+        """Returns SPEXone ApID.
         """
         if self.__hdr is None:
             return None
@@ -234,14 +191,15 @@ class CCSDSio:
 
     @property
     def grouping_flag(self) -> int:
-        """
-        Returns grouping flag
+        """Returns grouping flag
+
+        The meaning of the grouping flag values are::
 
-        Possible values:
           00 continuation packet-data segment
           01 first packet-data segment
           10 last packet-data segment
           11 packet-data unsegmented
+
         """
         if self.__hdr is None:
             return None
@@ -250,8 +208,7 @@ class CCSDSio:
 
     @property
     def sequence_count(self) -> int:
-        """
-        Returns sequence counter, rollover to zero at 0x3FFF
+        """Returns sequence counter, rollover to zero at 0x3FFF.
         """
         if self.__hdr is None:
             return None
@@ -260,11 +217,8 @@ class CCSDSio:
 
     @property
     def packet_length(self) -> int:
-        """
-        Returns size of packet data in bytes
+        """Returns size of packet data in bytes.
 
-        Notes
-        -----
         Value equals secondary header + user data (always odd)
         """
         if self.__hdr is None:
@@ -273,9 +227,8 @@ class CCSDSio:
         return self.__hdr['length']
 
     # ---------- define empty telemetry packet ----------
-    def open_next_file(self):
-        """
-        Open next file from file_list
+    def open_next_file(self) -> None:
+        """Open next file from file_list.
         """
         flname = next(self.file_list)
         if not Path(flname).is_file():
@@ -287,9 +240,18 @@ class CCSDSio:
 
     @staticmethod
     def fix_dem_hk24(dem_hk):
-        """
-        Correct 32-bit integers in the DemHK which originate from
-        24-bit integers in the detector register values
+        """Correct 32-bit integers in the DemHK which originate from
+        24-bit integers in the detector register values.
+
+        Parameters
+        ----------
+        dem_hk : numpy.ndarray
+           SPEXone DEM housekeeping packages
+
+        Returns
+        -------
+        numpy.ndarray
+           SPEXone DEM housekeeping packages
         """
         for key in ['DET_EXPTIME', 'DET_EXPSTEP',
                     'DET_KP1', 'DET_KP2', 'DET_EXPTIME2', 'DET_EXPSTEP2']:
@@ -299,13 +261,22 @@ class CCSDSio:
 
     @staticmethod
     def fix_sci_hk24(sci_hk):
-        """
-        Correct 32-bit integers in the Science HK which originate from
-        24-bit integers in the detector register values
+        """Correct 32-bit integers in the Science HK which originate from
+        24-bit integers in the detector register values.
 
         In addition:
          - copy the first 4 bytes of DET_CHENA to DET_ILVDS
          - parameter 'REG_BINNING_TABLE_START' was writen in little-endian
+
+        Parameters
+        ----------
+        sci_hk : numpy.ndarray
+           SPEXone Science telemetry packages
+
+        Returns
+        -------
+        numpy.ndarray
+           SPEXone Science telemetry packages
         """
         if np.all(sci_hk['ICUSWVER'] < 0x129):
             key = 'REG_BINNING_TABLE_START'
@@ -328,14 +299,167 @@ class CCSDSio:
 
         return sci_hk
 
-    def read_packet(self):
-        """
-        Read next telemetry packet
+    def __rd_science(self, hdr) -> np.ndarray:
+        """Read Science telemetry packet.
+
+        Parameters
+        ----------
+        hdr :  numpy.ndarray
+           CCSDS header information
 
         Returns
         -------
-        ndarray
+        numpy.ndarray
+           SPEXone Science telemetry packages
         """
+        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
+        packet = np.empty(1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('science_hk', SCIHK_DTYPE),
+            ('icu_time', TIME_DTYPE),
+            ('image_data', 'O')]))
+        packet['packet_header'] = hdr
+
+        # first segement or unsegmented data packet provides Science_HK
+        if self.grouping_flag in (1, 3):
+            packet['science_hk'] = self.fix_sci_hk24(
+                np.fromfile(self.fp, count=1, dtype=SCIHK_DTYPE))
+            num_bytes -= SCIHK_DTYPE.itemsize
+            packet['icu_time'] = np.fromfile(self.fp, count=1,
+                                             dtype=TIME_DTYPE)
+            num_bytes -= TIME_DTYPE.itemsize
+
+        # read detector image data
+        packet['image_data'][0] = np.fromfile(self.fp, dtype='>u2',
+                                              count=num_bytes // 2)
+        return packet
+
+    def __rd_nomhk(self, hdr) -> np.ndarray:
+        """Read NomHK telemetry packet.
+
+        Parameters
+        ----------
+        hdr :  numpy.ndarray
+           CCSDS header information
+
+        Returns
+        -------
+        numpy.ndarray
+           SPEXone Nominal housekeeping packages
+        """
+        packet = np.empty(1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('nominal_hk', tmtc_dtype(0x320))]))
+        packet['packet_header'] = hdr
+        packet['nominal_hk'] = np.fromfile(self.fp, count=1,
+                                           dtype=tmtc_dtype(0x320))
+        return packet
+
+    def __rd_demhk(self, hdr) -> np.ndarray:
+        """Read DemHK telemetry packet.
+
+        Parameters
+        ----------
+        hdr :  numpy.ndarray
+           CCSDS header information
+
+        Returns
+        -------
+        numpy.ndarray
+           SPEXone detector housekeeping packages
+        """
+        packet = np.empty(1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('detector_hk', tmtc_dtype(0x322))]))
+        packet['packet_header'] = hdr
+        packet['detector_hk'] = self.fix_dem_hk24(
+            np.fromfile(self.fp, count=1, dtype=tmtc_dtype(0x322)))
+        return packet
+
+    def __rd_tc_accept(self, _) -> np.ndarray:
+        """Read/dump TcAccept packet.
+        """
+        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
+        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('TcPacketId', '>u2'),
+            ('TcSeqControl', '>u2')]))
+        print('[TcAccept]: ', self, packet['TcPacketId'][0],
+              packet['TcSeqControl'][0])
+        return packet
+
+    def __rd_tc_execute(self, _) -> np.ndarray:
+        """Read/dump TcExecute packet.
+        """
+        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
+        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('TcPacketId', '>u2'),
+            ('TcSeqControl', '>u2')]))
+        print('[TcExecute]:', self, packet['TcPacketId'][0],
+              packet['TcSeqControl'][0])
+        return packet
+
+    def __rd_tc_fail(self, _) -> np.ndarray:
+        """Read/dump TcFail packet.
+        """
+        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
+        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('TcPacketId', '>u2'),
+            ('TcSeqControl', '>u2'),
+            ('TcFailCode', '>u2'),
+            ('FailParameter1', '>u2'),
+            ('FailParameter2', '>u2')]))
+        print('[TcFail]:   ', self, packet['TcPacketId'][0],
+              packet['TcSeqControl'][0], bin(packet['TcFailCode'][0]),
+              packet['FailParameter1'][0], packet['FailParameter2'][0]
+              )
+        return packet
+
+    def __rd_tc_reject(self, _) -> np.ndarray:
+        """Read/dump TcReject packet.
+        """
+        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
+        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('TcPacketId', '>u2'),
+            ('TcSeqControl', '>u2'),
+            ('TcRejectCode', '>u2'),
+            ('RejectParameter1', '>u2'),
+            ('RejectParameter2', '>u2')]))
+        print('[TcReject]: ', self, packet['TcPacketId'][0],
+              packet['TcSeqControl'][0], bin(packet['TcRejectCode'][0]),
+              packet['RejectParameter1'][0], packet['RejectParameter2'][0])
+        return packet
+
+    def __rd_other(self, hdr) -> np.ndarray:
+        """Read other telemetry packet.
+        """
+        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
+        if not 0x320 <= self.ap_id <= 0x350:
+            self.found_invalid_apid = True
+            self.fp.seek(num_bytes, 1)
+            return None
+
+        packet = np.empty(1, dtype=np.dtype([
+            ('packet_header', HDR_DTYPE),
+            ('raw_data', 'u1', num_bytes)]))
+        packet['packet_header'] = hdr
+        packet['raw_data'] = np.fromfile(self.fp, count=num_bytes, dtype='u1')
+        return packet
+
+    def read_packet(self) -> np.ndarray:
+        """Read next telemetry packet.
+
+        Returns
+        -------
+        numpy.ndarray
+           CCSDS packet data
+        """
+        if self.fp is None:
+            return None
+
         # read primary/secondary header
         hdr = np.fromfile(self.fp, count=1, dtype=HDR_DTYPE)
         if hdr.size == 0:
@@ -352,72 +476,32 @@ class CCSDSio:
 
         # save packet header as class attribute
         self.__hdr = hdr[0]
-        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
 
-        if self.ap_id == 0x350:             # Science telemetry packet
-            packet = np.empty(1, dtype=np.dtype([
-                ('packet_header', HDR_DTYPE),
-                ('science_hk', SCIHK_DTYPE),
-                ('icu_time', TIME_DTYPE),
-                ('image_data', 'O')]))
-            packet['packet_header'] = hdr
-
-            # first segement or unsegmented data packet provides Science_HK
-            if self.grouping_flag in (1, 3):
-                packet['science_hk'] = self.fix_sci_hk24(
-                    np.fromfile(self.fp, count=1, dtype=SCIHK_DTYPE))
-                num_bytes -= SCIHK_DTYPE.itemsize
-                packet['icu_time'] = np.fromfile(self.fp, count=1,
-                                                 dtype=TIME_DTYPE)
-                num_bytes -= TIME_DTYPE.itemsize
-
-            # read detector image data
-            packet['image_data'][0] = np.fromfile(self.fp, dtype='>u2',
-                                                  count=num_bytes // 2)
-        elif self.ap_id == 0x320:        # NomHK telemetry packet
-            packet = np.empty(1, dtype=np.dtype([
-                ('packet_header', HDR_DTYPE),
-                ('nominal_hk', tmtc_dtype(0x320))]))
-            packet['packet_header'] = hdr
-            packet['nominal_hk'] = np.fromfile(self.fp, count=1,
-                                               dtype=tmtc_dtype(0x320))
-        elif self.ap_id == 0x322:        # DemHK telemetry packet
-            packet = np.empty(1, dtype=np.dtype([
-                ('packet_header', HDR_DTYPE),
-                ('detector_hk', tmtc_dtype(0x322))]))
-            packet['packet_header'] = hdr
-            packet['detector_hk'] = self.fix_dem_hk24(
-                np.fromfile(self.fp, count=1, dtype=tmtc_dtype(0x322)))
-        else:
-            if not 0x320 <= self.ap_id <= 0x350:
-                self.found_invalid_apid = True
-
-            packet = np.empty(1, dtype=np.dtype([
-                ('packet_header', HDR_DTYPE),
-                ('raw_data', 'u1', num_bytes)]))
-            packet['packet_header'] = hdr
-            packet['raw_data'] = np.fromfile(self.fp, count=num_bytes,
-                                             dtype='u1')
-            # move to the next telemetry packet
-            # self.fp.seek(num_bytes, 1)
-
-        return packet
+        # read telemetry packet
+        rd_func = {0x350: self.__rd_science,
+                   0x320: self.__rd_nomhk,
+                   0x322: self.__rd_demhk,
+                   0x331: self.__rd_tc_accept,
+                   0x332: self.__rd_tc_reject,
+                   0x333: self.__rd_tc_execute,
+                   0x334: self.__rd_tc_fail}.get(self.ap_id, self.__rd_other)
+        return rd_func(hdr)
 
     @staticmethod
-    def select_tm(packets_in: tuple, ap_id: int):
-        """
-        Select telemetry packages on SPEXone ApID
+    def select_tm(packets_in: tuple, ap_id: int) -> tuple:
+        """Select telemetry packages on SPEXone ApID.
 
         Parameters
         ----------
         packets: tuple
            Tuple with mix of SPEXone telemetry packages
         ap_id: int
-          SPEXone ApID
+           SPEXone ApID
 
         Returns
         -------
-        Tuple with selected telemetry packages
+        tuple
+           selected telemetry packages
         """
         packets = ()
         for packet in packets_in:
@@ -429,9 +513,8 @@ class CCSDSio:
 
         return packets
 
-    def science_tm(self, packets_in: tuple):
-        """
-        Combine segmented Science telemetry packages
+    def science_tm(self, packets_in: tuple) -> tuple:
+        """Combine segmented Science telemetry packages.
 
         Parameters
         ----------
@@ -440,18 +523,18 @@ class CCSDSio:
 
         Returns
         -------
-        Tuple with unsegmented Science telemetry packages
+        tuple
+           unsegmented Science telemetry packages
         """
         # reject non-Science telemetry packages
         packets = self.select_tm(packets_in, 0x350)
         if not packets:
             return ()
 
-        # first telemetry package must have grouping flag equals 1
+        # check if grouping_flag of first segement equals 1
+        #   else reject all segments with grouping_flag != 1
         self.__hdr = packets[0]['packet_header']
         if self.grouping_flag != 1:
-            msg = ('[WARNING]: rejected first image because it is incomplete'
-                   ' - received only {:d} segments.')
             ii = 0
             for packet in packets:
                 self.__hdr = packet['packet_header']
@@ -459,25 +542,25 @@ class CCSDSio:
                     break
                 ii += 1
 
-            print(msg.format(ii))
+            print(f'[WARNING]: first frame incomplete - skipped {ii} segments')
             packets = packets[ii:]
             if not packets:
                 return ()
 
-        # last telemetry package must have grouping flag equals 2
+        # check if grouping_flag of last segement equals 2
+        #   else reject all segments after the last segment
+        #   with grouping_flag == 2
         self.__hdr = packets[-1]['packet_header']
         if self.grouping_flag != 2:
-            msg = ('[WARNING]: rejected last image because it is incomplete'
-                   ' - received only {:d} segments.')
-            ii = len(packets)
-            while ii > 0:
-                ii -= 1
-                self.__hdr = packets[ii]['packet_header']
+            ii = 0
+            for packet in packets:
+                self.__hdr = packet['packet_header']
                 if self.grouping_flag == 2:
                     break
+                ii += 1
 
-            print(msg.format(len(packets) - (ii + 1)))
-            packets = packets[:ii+1]
+            print(f'[WARNING]: last frame incomplete - rejected {ii} segments')
+            packets = packets[:-ii]
             if not packets:
                 return ()
 
@@ -486,13 +569,18 @@ class CCSDSio:
         prev_grp_flag = 2
         for packet in packets:
             grouping_flag = (packet['packet_header']['sequence'] >> 14) & 0x3
-
+            # print(prev_grp_flag, grouping_flag, len(res), offs,
+            #      packet['image_data'].size)
             # handle segmented data
-            if grouping_flag == 1:   # first segment
+            if grouping_flag == 1:       # first segment
                 # group_flag of previous package should be 2
                 if prev_grp_flag != 2:
-                    msg = ('corrupted segements - detected APID 1 after <> 2')
-                    raise RuntimeError(msg)
+                    if packet['image_data'].size in (3853, 7853):
+                        print(MSG_SKIP_FRAME)
+                        prev_grp_flag = 2
+                        offs = 0
+                    else:
+                        raise RuntimeError(MSG_CORRUPT_APID)
 
                 img_size = packet['science_hk']['IMRLEN'] // 2
                 rec_buff = np.empty(1, dtype=np.dtype([
@@ -508,12 +596,10 @@ class CCSDSio:
                 img_buff[offs:offs + packet['image_data'].size] = \
                     packet['image_data']
                 offs += packet['image_data'].size
-            else:                         # continuation or last segment
+            else:                        # continuation or last segment
                 # group_flag of previous package should be 0 or 1
                 if prev_grp_flag == 2:
-                    msg = ('corrupted segements - detected segement of'
-                           ' new image, however, previous not closed')
-                    raise RuntimeError(msg)
+                    raise RuntimeError(MSG_CORRUPT_FRAME)
 
                 img_buff[offs:offs + packet['image_data'].size] = \
                     packet['image_data']
@@ -523,8 +609,7 @@ class CCSDSio:
                         rec_buff['image_data'] = img_buff
                         res += (rec_buff,)
                     else:
-                        print('[WARNING]: rejected image'
-                              ' because it is incomplete')
+                        print(MSG_SKIP_FRAME)
                     offs = 0
 
             # keep current group flag for next read
